# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
import PGUtils
from Game import GameState

MAX_VALUE = 6

def valid_conversions(grid, nps):
    # subset must be 2 or more and add up to less than 6
    for np in PGUtils.powerset(nps):
        if len(np)>=2:
            nsum = sum(abs(grid[p]) for p in np)
            if nsum<=MAX_VALUE: 
                yield np, nsum
                
class CephalopodGameState(GameState):
    def successors(self, ai=False):
        pnum = self.current_player
        dirs = GridUtils.ortho_direction_array(len(self.dimensions()))
        for c in self.enumerate_coords_with_value(0):
            neighbors = [GridUtils.add_tuples(c,d) for d in dirs]
            neighbors = [x for x in neighbors if self.valid_coords(x) and self.grid[x]!=0]
            found = False
            if len(neighbors)>1:
                for npositions, nsum in valid_conversions(self.grid, neighbors):
                    found = True
                    nextgrid = self.grid.copy()
                    # remove neighbors and put sum in center
                    for np in npositions:
                        nextgrid[np] = 0
                    nextgrid[c] = nsum * (1 if pnum==1 else -1)
                    nextstate = self.next_state_with_grid(nextgrid)
                    # for non-ai, need all permutations
                    nperms = [npositions] if ai else PGUtils.permutations(npositions)
                    for p in nperms:
                        yield (tuple([c]+p+[c]), nextstate)                    
            if not found:
                # drop a 1
                nextgrid = self.grid.copy()
                nextgrid[c] = (1 if pnum==1 else -1)
                yield ( (c,), self.next_state_with_grid(nextgrid))
    
    def successors_for_ai(self):
        return self.successors(ai=True)
        
    def is_game_over(self):
        return 0 not in self.grid
        
    def player_score(self, pnum):
        # 1 point for each die, regardless of value
        return sum(1 for val in self.itervalues() if (val>0 and pnum==1) or (val<0 and pnum==2))
        
    def animations_for_move(self, move, newstate):
        # when dice are removed, they should "shrink" into the new die that is absorbing their pips
        import PGPieceAnimation
        anim_group = GameState.animations_for_move(self, move, newstate)
        for anim in anim_group.animations:
            seg = anim.segments[0]
            if seg.action==PGPieceAnimation.REMOVE:
                seg.to_position = move[0]
        return anim_group


        